@@ -97,7 +97,7 @@ void SystemD::stop()
     try {
         manager->stopUnit(appId_, job_, instance_);
     } catch (const std::runtime_error& e) {
-        g_warning("Error occured while stopping unit: %s", e.what());
+        g_warning("Error occurred while stopping unit: %s", e.what());
     }
 }
 
@@ -114,6 +114,7 @@ static const char* SYSTEMD_DBUS_IFACE_SERVICE{"org.freedesktop.systemd1.Service"
 
 SystemD::SystemD(const std::shared_ptr<Registry::Impl>& registry)
     : Base(registry)
+    , handle_reloading(DBusSignalUnsubscriber{})
     , handle_unitNew(DBusSignalUnsubscriber{})
     , handle_unitRemoved(DBusSignalUnsubscriber{})
     , handle_appFailed(DBusSignalUnsubscriber{})
@@ -203,6 +204,35 @@ void SystemD::setupUserbus(const std::shared_ptr<Registry::Impl>& reg)
                                },
                                nullptr);
 
+        /* Setup reloading signal first to avoid race-condition. */
+        handle_reloading = managedDBusSignalConnection(
+            g_dbus_connection_signal_subscribe(bus.get(),                  /* bus */
+                                               nullptr,                    /* sender */
+                                               SYSTEMD_DBUS_IFACE_MANAGER, /* interface */
+                                               "Reloading",                /* signal */
+                                               SYSTEMD_DBUS_PATH_MANAGER,  /* path */
+                                               nullptr,                    /* arg0 */
+                                               G_DBUS_SIGNAL_FLAGS_NONE,
+                                               [](GDBusConnection*, const gchar*, const gchar*, const gchar*,
+                                                  const gchar*, GVariant* params, gpointer user_data) -> void {
+                                                   auto pthis = static_cast<SystemD*>(user_data);
+
+                                                   if (!g_variant_check_format_string(params, "(b)", FALSE))
+                                                   {
+                                                       g_warning("Got 'Reloading' signal with unknown parameter type: %s",
+                                                                 g_variant_get_type_string(params));
+                                                       return;
+                                                   }
+
+                                                   gboolean active{false};
+                                                   g_variant_get(params, "(b)", &active);
+
+                                                   pthis->systemdReloading(active);
+                                               },        /* callback */
+                                               this,     /* user data */
+                                               nullptr), /* user data destroy */
+            bus);
+
         /* Setup Unit add/remove signals */
         handle_unitNew = managedDBusSignalConnection(
             g_dbus_connection_signal_subscribe(bus.get(),                  /* bus */
@@ -248,8 +278,7 @@ void SystemD::setupUserbus(const std::shared_ptr<Registry::Impl>& reg)
 
                                                    try
                                                    {
-                                                       auto info = pthis->unitNew(unitname, unitpath, pthis->userbus_);
-                                                       pthis->sig_jobStarted(info.job, info.appid, info.inst);
+                                                       pthis->unitNew(unitname, unitpath, pthis->userbus_);
                                                    }
                                                    catch (std::runtime_error& e)
                                                    {
@@ -708,6 +737,14 @@ std::shared_ptr<Application::Instance> SystemD::launch(
         g_variant_builder_close(&builder);
         g_variant_builder_close(&builder);
 
+        /* CollectMode */
+        g_variant_builder_open(&builder, G_VARIANT_TYPE_TUPLE);
+        g_variant_builder_add_value(&builder, g_variant_new_string("CollectMode"));
+        g_variant_builder_open(&builder, G_VARIANT_TYPE_VARIANT);
+        g_variant_builder_add_value(&builder, g_variant_new_string("inactive-or-failed"));
+        g_variant_builder_close(&builder);
+        g_variant_builder_close(&builder);
+
         /* Type */
         g_variant_builder_open(&builder, G_VARIANT_TYPE_TUPLE);
         g_variant_builder_add_value(&builder, g_variant_new_string("Type"));
@@ -729,7 +766,7 @@ std::shared_ptr<Application::Instance> SystemD::launch(
 
         /* Clean up env before shipping it */
         for (const auto& rmenv :
-             {"APP_DIR", "APP_URIS", "APP_EXEC", "APP_EXEC_POLICY", "APP_LAUNCHER_PID",
+             {"APP_URIS", "APP_EXEC", "APP_EXEC_POLICY", "APP_LAUNCHER_PID",
               "INSTANCE_ID", "MIR_SERVER_PLATFORM_PATH", "MIR_SERVER_PROMPT_FILE", "MIR_SERVER_HOST_SOCKET",
               "LOMIRI_APP_LAUNCH_OOM_HELPER", "LOMIRI_APP_LAUNCH_LEGACY_ROOT"})
         {
@@ -895,9 +932,32 @@ std::string SystemD::unitPath(const SystemD::UnitInfo& info)
     return reg->thread.executeOnThread<std::string>([&data]() { return data->unitpath; });
 }
 
-SystemD::UnitInfo SystemD::unitNew(const std::string& name,
-                                   const std::string& path,
-                                   const std::shared_ptr<GDBusConnection>& bus)
+void SystemD::systemdReloading(bool reloading)
+{
+    bool wasReloading = systemdIsReloading_;
+    systemdIsReloading_ = reloading;
+
+    if (wasReloading && !reloading) {
+        /* Handle UnitRemoved without corresponding UnitNew. */
+        auto it = unitPaths.begin();
+        while (it != unitPaths.end()) {
+            if (it->second->pendingRemoval)
+            {
+                auto info = it->first;
+                it = unitPaths.erase(it);
+                sig_jobStopped(info.job, info.appid, info.inst);
+            }
+            else
+            {
+                it++;
+            }
+        }
+    }
+}
+
+void SystemD::unitNew(const std::string& name,
+                      const std::string& path,
+                      const std::shared_ptr<GDBusConnection>& bus)
 {
     if (path == "/")
     {
@@ -913,10 +973,18 @@ SystemD::UnitInfo SystemD::unitNew(const std::string& name,
     auto data = std::make_shared<UnitData>();
     data->jobpath = path;
 
-    /* We already have this one, continue on */
-    if (!unitPaths.insert(std::make_pair(info, data)).second)
+    auto insertResult = unitPaths.insert(std::make_pair(info, data));
+    if (!insertResult.second)
     {
-        throw std::runtime_error{"Duplicate unit, not really new"};
+        if (insertResult.first->second->pendingRemoval)
+        {
+            insertResult.first->second->pendingRemoval = false;
+            return;
+        }
+        else
+        {
+            throw std::runtime_error{"Duplicate unit, not really new"};
+        }
     }
 
     /* We need to get the path, we're blocking everyone else on
@@ -952,7 +1020,7 @@ SystemD::UnitInfo SystemD::unitNew(const std::string& name,
         data->unitpath = gpath;
     }
 
-    return info;
+    sig_jobStarted(info.job, info.appid, info.inst);
 }
 
 void SystemD::unitRemoved(const std::string& name, const std::string& path)
@@ -960,7 +1028,14 @@ void SystemD::unitRemoved(const std::string& name, const std::string& path)
     UnitInfo info = parseUnit(name);
 
     auto it = unitPaths.find(info);
-    if (it != unitPaths.end())
+    if (it == unitPaths.end())
+        return;
+
+    if (systemdIsReloading_)
+    {
+        it->second->pendingRemoval = true;
+    }
+    else
     {
         unitPaths.erase(it);
         sig_jobStopped(info.job, info.appid, info.inst);
